@@ -8,10 +8,11 @@ import copy
 import time
 
 from mnemosyne.libmnemosyne.fact import Fact
-from mnemosyne.libmnemosyne.translator import _
+from mnemosyne.libmnemosyne.gui_translator import _
 from mnemosyne.libmnemosyne.controller import Controller
 from mnemosyne.libmnemosyne.utils import remove_empty_dirs_in
 from mnemosyne.libmnemosyne.utils import expand_path, contract_path
+from mnemosyne.libmnemosyne.card_type_converter import CardTypeConverter
 
 HOUR = 60 * 60 # Seconds in an hour.
 DAY = 24 * HOUR # Seconds in a day.
@@ -46,7 +47,12 @@ class DefaultController(Controller):
 
         """
 
+        # FIXME: For some reason, datetime.datetime.now() does not respect
+        # DST here, although it does work in a standalone Python program...
+
         if time.time() > self.next_rollover:
+            for f in self.component_manager.all("hook", "at_rollover"):
+                f.run()
             if self.config().server_only:
                 self.database().backup()
                 self.log().dump_to_science_log()
@@ -66,7 +72,12 @@ class DefaultController(Controller):
                 self.log().activate()
                 self.config().save()
                 self.reset_study_mode()
-            self.next_rollover = self.database().start_of_day_n_days_ago(n=-1)
+            previous_rollover = self.next_rollover
+            next_rollover = self.database().start_of_day_n_days_ago(n=-1)
+            # Avoid rare issue with DST.
+            if abs(next_rollover - previous_rollover) < HOUR:
+                next_rollover += DAY
+            self.next_rollover = next_rollover
         if db_maintenance and \
            (time.time() > self.config()["last_db_maintenance"] + 90 * DAY):
             self.component_manager.current("database_maintenance").run()
@@ -134,9 +145,9 @@ class DefaultController(Controller):
                 if partial_tag:
                     partial_tag += "::"
                 partial_tag += node
-                if  partial_tag in tag_names:
+                if partial_tag in tag_names:
                     parent_tag_names.append(partial_tag)
-        return list(set(tag_names) - set(parent_tag_names))
+        return sorted(list(set(tag_names) - set(parent_tag_names)))
 
     def create_new_cards(self, fact_data, card_type, grade, tag_names,
                          check_for_duplicates=True, save=True):
@@ -287,14 +298,17 @@ class DefaultController(Controller):
         _("Card data not correctly formatted for conversion.\n\nSkipping ") +\
                 "|".join(list(fact.data.values())) + ".\n")
             return -2
-        converter = self.component_manager.current\
-              ("card_type_converter", used_for=(old_card_type.__class__,
-                                                new_card_type.__class__))
+        # For conversion, we need to look at the top ancestor.
+        ancestor_id_old = old_card_type.id.split("::", maxsplit=1)[0]
+        ancestor_id_new = new_card_type.id.split("::", maxsplit=1)[0]
+        converter = None
+        if ancestor_id_old != ancestor_id_new:
+            converter = self.component_manager.current\
+              ("card_type_converter", CardTypeConverter.card_type_converter_key\
+               (self.card_type_with_id(ancestor_id_old),
+                self.card_type_with_id(ancestor_id_new)))
         if not converter:
-            # Perhaps they have a common ancestor.
-            parents_old = old_card_type.id.split("::")
-            parents_new = new_card_type.id.split("::")
-            if parents_old[0] == parents_new[0]:
+            if ancestor_id_old == ancestor_id_new:
                 edited_cards = cards_from_fact
                 new_fact_view_for = {}
                 for index, old_fact_view in enumerate(old_card_type.fact_views):
@@ -463,6 +477,9 @@ _("This card has different tags than its sister cards. Update tags for current c
         w.close_progress()
 
     def star_current_card(self):
+        review_controller = self.review_controller()
+        if not review_controller.card:
+            return
         self.stopwatch().pause()
         self.flush_sync_server()
         if self.config()["star_help_shown"] == False:
@@ -470,7 +487,6 @@ _("This card has different tags than its sister cards. Update tags for current c
 _("This will add a tag 'Starred' to the current card, so that you can find it back easily, e.g. to edit it on a desktop."))
             self.config()["star_help_shown"] = True
         db = self.database()
-        review_controller = self.review_controller()
         tag = db.get_or_create_tag_with_name(_("Starred"))
         _sister_card_ids = [card._id for card in \
             self.database().cards_from_fact(review_controller.card.fact)]
@@ -481,10 +497,12 @@ _("This will add a tag 'Starred' to the current card, so that you can find it ba
         self.stopwatch().unpause()
 
     def delete_current_card(self):
+        review_controller = self.review_controller()
+        if not review_controller.card:
+            return
         self.stopwatch().pause()
         self.flush_sync_server()
         db = self.database()
-        review_controller = self.review_controller()
         fact = review_controller.card.fact
         no_of_cards = len(db.cards_from_fact(fact))
         if no_of_cards == 1:
@@ -706,7 +724,8 @@ _("Your database will be autosaved before exiting. Also, it is saved every coupl
             self.config()["single_database_help_shown"] = True
         self.flush_sync_server()
         suffix = self.database().suffix
-        old_path = expand_path(self.config()["last_database"], self.config().data_dir)
+        old_path = expand_path(self.config()["last_database"],
+                               self.config().data_dir)
         old_media_dir = self.database().media_dir()
         filename = self.main_widget().get_filename_to_save(path=old_path,
             filter=_("Mnemosyne databases") + " (*%s)" % suffix)
@@ -841,6 +860,7 @@ _("The configuration database cannot be used to store cards."))
     def show_activate_cards_dialog_post(self):
         review_controller = self.review_controller()
         review_controller.reset_but_try_to_keep_current_card()
+        review_controller.reload_counters()
         review_controller.update_status_bar_counters()
         self.update_title()
         self.stopwatch().unpause()
@@ -855,6 +875,7 @@ _("This will tag all the cards in a given card type which have the same question
         self.database().tag_all_duplicates()
         review_controller = self.review_controller()
         review_controller.reset_but_try_to_keep_current_card()
+        review_controller.reload_counters()
         review_controller.update_status_bar_counters()
         self.stopwatch().unpause()
 
