@@ -6,7 +6,6 @@ import os
 import sys
 import time
 import string
-import shutil
 import datetime
 import copy as objcopy
 
@@ -15,13 +14,13 @@ from openSM2sync.log_entry import EventTypes
 from mnemosyne.libmnemosyne.tag import Tag
 from mnemosyne.libmnemosyne.fact import Fact
 from mnemosyne.libmnemosyne.card import Card
-from mnemosyne.libmnemosyne.translator import _
+from mnemosyne.libmnemosyne.gui_translator import _
 from mnemosyne.libmnemosyne.database import Database
 from mnemosyne.libmnemosyne.card_type import CardType
 from mnemosyne.libmnemosyne.fact_view import FactView
 from mnemosyne.libmnemosyne.utils import traceback_string, copy
-from mnemosyne.libmnemosyne.utils import numeric_string_cmp, mangle
 from mnemosyne.libmnemosyne.utils import expand_path, contract_path
+from mnemosyne.libmnemosyne.utils import numeric_string_cmp_key, mangle
 
 # All ids beginning with an underscore refer to primary keys in the SQL
 # database. All other id's correspond to the id's used in libmnemosyne.
@@ -205,6 +204,7 @@ from mnemosyne.libmnemosyne.databases.SQLite_media import SQLiteMedia
 from mnemosyne.libmnemosyne.databases.SQLite_logging import SQLiteLogging
 from mnemosyne.libmnemosyne.databases.SQLite_statistics import SQLiteStatistics
 
+
 class SQLite(Database, SQLiteSync, SQLiteMedia, SQLiteLogging,
              SQLiteStatistics):
 
@@ -220,7 +220,7 @@ class SQLite(Database, SQLiteSync, SQLiteMedia, SQLiteLogging,
 
     """
 
-    version = "2"
+    version = "4"
     suffix = ".db"
     store_pregenerated_data = True
 
@@ -271,9 +271,9 @@ class SQLite(Database, SQLiteSync, SQLiteMedia, SQLiteLogging,
 
     def path(self):
         return self._path
-    
+
     def data_dir(self):
-        return os.path.dirname(self._path)    
+        return os.path.dirname(self._path)
 
     def name(self):
         return os.path.basename(self._path)
@@ -289,7 +289,7 @@ class SQLite(Database, SQLiteSync, SQLiteMedia, SQLiteLogging,
         self.main_widget().set_progress_text(_("Defragmenting database..."))
         self.con.execute("vacuum")
         # Make sure the "Untagged" tag does not show up together with
-        # different tags.
+        # different tags (not sure if bug causing this has been fixed).
         untagged = self.tag("__UNTAGGED__", is_id_internal=False)
         for cursor in self.con.execute("select _id from cards"):
             _card_id = cursor[0]
@@ -300,6 +300,9 @@ class SQLite(Database, SQLiteSync, SQLiteMedia, SQLiteLogging,
                 self.con.execute(\
                     "delete from tags_for_card where _card_id=? and _tag_id=?",
                     (_card_id, untagged._id))
+        # Make sure no orphaned card tags exist (not sure if bug causing
+        # this has been fixed).
+        self.con.execute("delete from tags_for_card where _card_id is null")
         self.main_widget().close_progress()
 
     def new(self, path):
@@ -334,21 +337,22 @@ class SQLite(Database, SQLiteSync, SQLiteMedia, SQLiteLogging,
         self._current_criterion.name = self.default_criterion_name
         self._current_criterion._tag_ids_active.add(tag._id)
         self.add_criterion(self._current_criterion)
+        self.save()
 
     def load(self, path):
         if self.is_loaded():
             self.unload()
         self._path = expand_path(path, self.config().data_dir)
         if not os.path.exists(self._path):
-            return new(path)
+            return self.new(path)
         # Check database version.
         try:
             sql_res = self.con.execute("""select value from global_variables
                 where key=?""", ("version", )).fetchone()
         except:
-            raise RuntimeError, _("Unable to load file at") + " " + self._path
+            raise RuntimeError(_("Unable to load file at") + " " + self._path)
         if sql_res is None:
-            raise RuntimeError, _("Unable to load file, query failed.")
+            raise RuntimeError(_("Unable to load file, query failed."))
         if sql_res[0] != self.version:
             if sql_res[0] == "Mnemosyne SQL 1.0":
                 previous_version = 1
@@ -360,7 +364,7 @@ class SQLite(Database, SQLiteSync, SQLiteMedia, SQLiteLogging,
                         import Upgrade2
                     Upgrade2(self.component_manager).run()
             except:
-                raise RuntimeError, _("Database upgrade failed.")
+                raise RuntimeError(_("Database upgrade failed."))
         self.create_media_dir_if_needed()
         # Upgrade.
         self.con.execute("""create index if not exists
@@ -372,13 +376,13 @@ class SQLite(Database, SQLiteSync, SQLiteMedia, SQLiteLogging,
             used_ids = [cursor[0] for cursor in \
                 self.con.execute("select distinct card_type_id from cards")]
         except:
-            raise RuntimeError, _("Unable to load file.") + traceback_string()
+            raise RuntimeError(_("Unable to load file.") + traceback_string())
         # We also need to do this for the card types which are only defined
         # and have no cards yet.
         defined_in_database_ids = [cursor[0] for cursor in \
             self.con.execute("select id from card_types")]
         for id in used_ids + defined_in_database_ids:
-            self.activate_plugins_for_card_type_with_id(id) 
+            self.activate_plugins_for_card_type_with_id(id)
         # Instantiate card types stored in this database. Since they could
         # depend on a plugin, the card types need to be instatiated last.
         for id in defined_in_database_ids:
@@ -391,7 +395,7 @@ class SQLite(Database, SQLiteSync, SQLiteMedia, SQLiteLogging,
         for f in self.component_manager.all("hook", "after_load"):
             f.run()
         # We don't log the database load here, but in libmnemosyne.__init__,
-        # as we prefer to log the start of the program first.                          
+        # as we prefer to log the start of the program first.
 
     def save(self, path=None):
         # Update format.
@@ -406,9 +410,9 @@ class SQLite(Database, SQLiteSync, SQLiteMedia, SQLiteLogging,
             if sys.platform == "win32":  # pragma: no cover
                 drive = os.path.splitdrive(path)[0]
                 import ctypes
-                if ctypes.windll.kernel32.GetDriveTypeW(u"%s\\" % drive) == 4:
-                    raise RuntimeError, \
-_("Putting a database on a network drive is forbidden under Windows to avoid data corruption.")
+                if ctypes.windll.kernel32.GetDriveTypeW("%s\\" % drive) == 4:
+                    raise RuntimeError(\
+_("Putting a database on a network drive is forbidden under Windows to avoid data corruption."))
             copy(self._path, dest_path)
             self._path = dest_path
         self.config()["last_database"] \
@@ -418,12 +422,16 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
 
     def backup(self):
         self.save()
-        if self.config()["backups_to_keep"] == 0:
+        if self.config()["max_backups"] == 0:
             return
         backupdir = os.path.join(self.config().data_dir, "backups")
         db_name = os.path.basename(self._path).rsplit(".", 1)[0]
-        backupfile = db_name + "-" + \
-            datetime.datetime.today().strftime("%Y%m%d-%H%M%S.db")
+        try:
+            backupfile = db_name + "-" + \
+                datetime.datetime.today().strftime("%Y%m%d-%H%M%S.db")
+        except:  # Work around strange Android library bug.
+            from mnemosyne.libmnemosyne.utils import rand_uuid
+            backupfile = db_name + "-" + rand_uuid() + ".db"
         backupfile = os.path.join(backupdir, backupfile)
         failed = False
         try:
@@ -438,13 +446,11 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
         for f in self.component_manager.all("hook", "after_backup"):
             f.run(backupfile)
         # Only keep the last logs.
-        if self.config()["backups_to_keep"] < 0:
-            return backupfile
-        files = [f for f in os.listdir(unicode(backupdir)) \
+        files = [f for f in os.listdir(backupdir) \
                 if f.startswith(db_name + "-")]
         files.sort()
-        if len(files) > self.config()["backups_to_keep"]:
-            surplus = len(files) - self.config()["backups_to_keep"]
+        if len(files) > self.config()["max_backups"]:
+            surplus = len(files) - self.config()["max_backups"]
             for file in files[0:surplus]:
                 os.remove(os.path.join(backupdir, file))
         return backupfile
@@ -477,7 +483,7 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
             self.log().dump_to_science_log()
             self.backup()  # Saves too.
             self._connection.close()
-        except Exception, e:
+        except Exception as e:
             pass
         finally:
             self._connection = None
@@ -601,7 +607,7 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
                 answer = self.main_widget().show_question(\
                     _("Make tag '%s' active in saved set '%s'?") % \
                     (tag.name, saved_criterion.name), _("Yes"), _("No"), "")
-            except NotImplementedError: 
+            except NotImplementedError:
                 # We are running in a non interactive mode.
                 answer = 0  # Yes
             if answer == 1:  # No.
@@ -638,8 +644,8 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
         new_name = tag.name
         stored_name = self.con.execute("select name from tags where _id=?",
             (tag._id, )).fetchone()[0]
-        if new_name != stored_name and self.con.execute("""select count() from
-            tags where name=?""", (new_name, )).fetchone()[0] != 0:
+        if new_name != stored_name and self.con.execute("""select 1 from
+            tags where name=? limit 1""", (new_name, )).fetchone() is not None:
             _existing_tag_id = self.con.execute("""select _id from tags where
             name=?""", (new_name, )).fetchone()[0]
             _card_ids_affected = [cursor[0] for cursor in self.con.execute(\
@@ -648,9 +654,9 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
             for _card_id in _card_ids_affected:
                 # If the card already had a tag with the updated name, delete
                 # the other tag.
-                if self.con.execute("""select count() from tags_for_card where
-                    _tag_id=? and _card_id=?""", (_existing_tag_id, _card_id))\
-                    .fetchone()[0] > 0:
+                if self.con.execute("""select 1 from tags_for_card where
+                    _tag_id=? and _card_id=? limit 1""",
+                    (_existing_tag_id, _card_id)).fetchone() is not None:
                     self.con.execute("""delete from tags_for_card where
                     _tag_id=? and _card_id=?""", (tag._id, _card_id))
                 # If not, update the link.
@@ -691,7 +697,7 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
                     (cursor[0], )).fetchone()[0]
                 if tag_name != "__UNTAGGED__":
                     tag_names.append(tag_name)
-            sorted_tag_names = sorted(tag_names, cmp=numeric_string_cmp)
+            sorted_tag_names = sorted(tag_names, key=numeric_string_cmp_key)
             tag_string = ", ".join(sorted_tag_names)
             self.con.execute("update cards set tags=? where _id=?",
                 (tag_string, _card_id))
@@ -706,8 +712,8 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
         self.con.execute("delete from tags_for_card where _tag_id=?",
             (tag._id, ))
         for _card_id in _card_ids_affected:
-            if self.con.execute("""select count() from tags_for_card where
-                _card_id=?""", (_card_id, )).fetchone()[0] == 0:
+            if self.con.execute("""select 1 from tags_for_card where
+                _card_id=? limit 1""", (_card_id, )).fetchone() is None:
                 untagged = self.get_or_create_tag_with_name("__UNTAGGED__")
                 self.con.execute("""insert into tags_for_card(_tag_id,
                     _card_id) values(?,?)""", (untagged._id, _card_id))
@@ -736,9 +742,9 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
     def delete_tag_if_unused(self, tag):
         if tag.id == "__UNTAGGED__":
             return
-        if self.con.execute("""select count() from tags as cat,
+        if self.con.execute("""select 1 from tags as cat,
             tags_for_card as cat_c where cat_c._tag_id=cat._id and
-            cat._id=?""", (tag._id, )).fetchone()[0] == 0:
+            cat._id=? limit 1""", (tag._id, )).fetchone() is None:
             self.delete_tag(tag)
 
     def tags(self):
@@ -749,9 +755,9 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
 
         result = [self.tag(cursor[0], is_id_internal=True) for cursor in \
             self.con.execute("select _id from tags")]
-        result.sort(key=lambda x: x.name, cmp=numeric_string_cmp)
+        result.sort(key=lambda x: numeric_string_cmp_key(x.name))
         index = 0
-        # __UNTAGGED__ is typically at the head of the list, apart when tags
+        # __UNTAGGED__ is typically at the head of the list, except when tags
         # start with numbers.
         for tag in result:
             if tag.name == "__UNTAGGED__":
@@ -760,10 +766,10 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
                 break
             index += 1
         return result
-    
+
     def has_tag_with_id(self, id):
-        return self.con.execute("select count() from tags where id=?",
-            (id, )).fetchone()[0] != 0
+        return self.con.execute("select 1 from tags where id=? limit 1",
+            (id, )).fetchone() is not None
 
     #
     # Facts.
@@ -815,10 +821,33 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
             (fact._id, ))
         self.log().deleted_fact(fact)
         del fact
-        
+
     def has_fact_with_id(self, id):
-        return self.con.execute("select count() from facts where id=?",
-            (id, )).fetchone()[0] != 0        
+        return self.con.execute("select 1 from facts where id=? limit 1",
+            (id, )).fetchone() is not None
+
+    def fact_ids_forgotten_and_learned_today(self, start_of_day, end_of_day):
+        return (cursor[0] for cursor in self.con.execute(
+            """
+            select cards._fact_id from log inner join cards where
+            log.object_id = cards.id and log.timestamp >= :start_of_day and
+            log.timestamp < :end_of_day and log.event_type = :event_type and
+            log.grade >= 2 and log.object_id in (
+              select object_id from log where timestamp >= :start_of_day and
+              timestamp < :end_of_day and event_type = :event_type and
+              grade < 2 and ret_reps > 0 group by object_id)
+            group by log.object_id
+            """,
+            {"start_of_day": start_of_day,
+             "end_of_day": end_of_day,
+             "event_type": EventTypes.REPETITION}).fetchall())
+
+    def fact_ids_newly_learned_today(self, start_of_day, end_of_day):
+        return (cursor[0] for cursor in self.con.execute(
+            """select cards._fact_id from log inner join cards where
+            log.object_id = cards.id and ?<=log.timestamp and log.timestamp<?
+            and log.event_type=? and log.grade>=2 and log.ret_reps==0""",
+            (start_of_day, end_of_day, EventTypes.REPETITION)).fetchall())
 
     #
     # Cards.
@@ -941,7 +970,7 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
                 _card_id) values(?,?)""", (tag._id, card._id))
 
     def delete_card(self, card, check_for_unused_tags=True):
-        if card._id is None: 
+        if card._id is None:
             # A card which was created and deleted before a sync, so that
             # it has incomplete information.
             self.con.execute("delete from cards where id=?", (card.id, ))
@@ -987,6 +1016,13 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
             values(?,?)""", arguments)
         if self.store_pregenerated_data:
             self._update_tag_strings(_card_ids)
+        # Apply criterion. (There does not seem to be any watertight shortcut
+        # we can take for a special case, especially when a card can have many
+        # tags, so we apply the criterion in full.)
+        criterion = self.current_criterion()
+        applier = self.component_manager.current("criterion_applier",
+            used_for=criterion.__class__)
+        applier.apply_to_database(criterion)
         # We don't call 'self.log.edited_card(card)', which would require us to
         # construct the entire card object, but take a short cut.
         for _card_id in _card_ids:
@@ -1014,6 +1050,13 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
         self.delete_tag_if_unused(tag)
         if self.store_pregenerated_data:
             self._update_tag_strings(_card_ids)
+        # Apply criterion. (There does not seem to be any watertight shortcut
+        # we can take for a special case, especially when a card can have many
+        # tags, so we apply the criterion in full.)
+        criterion = self.current_criterion()
+        applier = self.component_manager.current("criterion_applier",
+            used_for=criterion.__class__)
+        applier.apply_to_database(criterion)
         # We don't call 'self.log.edited_card(card)', which would require us
         # to construct the entire card object, but take a short cut.
         for _card_id in _card_ids:
@@ -1024,9 +1067,9 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
                 (EventTypes.EDITED_CARD, int(time.time()), card_id))
 
     def has_card_with_id(self, id):
-        return self.con.execute("select count() from cards where id=?",
-            (id, )).fetchone()[0] != 0 
-    
+        return self.con.execute("select 1 from cards where id=? limit 1",
+            (id, )).fetchone() is not None
+
     #
     # Fact views.
     #
@@ -1077,19 +1120,19 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
             (fact_view.id, ))
         self.log().deleted_fact_view(fact_view)
         del fact_view
-        
+
     def has_fact_view_with_id(self, id):
-        return self.con.execute("select count() from fact_views where id=?",
-            (id, )).fetchone()[0] != 0         
+        return self.con.execute("select 1 from fact_views where id=? limit 1",
+            (id, )).fetchone() is not None
 
     #
     # Card types.
     #
-    
+
     def activate_plugins_for_card_type_with_id(self, id):
         builtin_ids = set(card_type.id for card_type in self.card_types())
         defined_in_database_ids = [cursor[0] for cursor in \
-            self.con.execute("select id from card_types")]        
+            self.con.execute("select id from card_types")]
         # Check if parents are missing plugins.
         plugin_needed_ids = set()
         while "::" in id: # Move up one level of the hierarchy.
@@ -1108,13 +1151,14 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
                         try:
                             plugin.activate()
                         except:
-                            raise RuntimeError, _("Error when running plugin:") \
-                                + "\n" + traceback_string()
+                            raise RuntimeError(_("Error when running plugin:") \
+                                + "\n" + traceback_string())
             if not found:
-                raise RuntimeError, \
-                _("Missing plugin for card type with id:") + " " + card_type_id      
+                raise RuntimeError(_("Missing plugin for card type with id:") +\
+                                   " " + card_type_id)
 
     def add_card_type(self, card_type):
+        card_type.extra_data["hidden_from_UI"] = card_type.hidden_from_UI
         self.con.execute("""insert into card_types(id, name,
             fact_keys_and_names, unique_fact_keys, required_fact_keys,
             fact_view_ids, keyboard_shortcuts, extra_data)
@@ -1147,8 +1191,8 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
             criteria_to_activate_card_type_in = [current_criterion]
         else:
             answer = self.main_widget().show_question(\
-                _("Make new card type active in saved set '%s'?") % \
-                (saved_criterion.name,), _("Yes"), _("No"), "")
+                _("Make new card type '%s' active in saved set '%s'?") % \
+                (card_type.name, saved_criterion.name), _("Yes"), _("No"), "")
             if answer == 1:  # No.
                 criteria_to_activate_card_type_in = []
             else:
@@ -1183,23 +1227,25 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
         card_type.required_fact_keys = eval(sql_res[3])
         card_type.keyboard_shortcuts = eval(sql_res[5])
         self._construct_extra_data(sql_res[6], card_type)
+        if "hidden_from_UI" in card_type.extra_data:
+            card_type.hidden_from_UI = card_type.extra_data["hidden_from_UI"]
         card_type.fact_views = [self.fact_view(fact_view_id,
             is_id_internal=False) for fact_view_id in eval(sql_res[4])]
         return card_type
 
     def is_user_card_type(self, card_type):
-        return self.con.execute("select count() from card_types where id=?",
-            (card_type.id, )).fetchone()[0] == 1
+        return self.con.execute("select 1 from card_types where id=? limit 1",
+            (card_type.id, )).fetchone() is not None
 
     def is_in_use(self, card_type):
         return self.con.execute(\
-            "select count() from cards where card_type_id=?",
-            (card_type.id, )).fetchone()[0] != 0
+            "select 1 from cards where card_type_id=? limit 1",
+            (card_type.id, )).fetchone() is not None
 
     def has_clones(self, card_type):
-         return self.con.execute(\
-            "select count() from card_types where id like ?",
-            (card_type.id + "::%", )).fetchone()[0] != 0
+        return self.con.execute(\
+            "select 1 from card_types where id like ? limit 1",
+            (card_type.id + "::%", )).fetchone() is not None
 
     def update_card_type(self, card_type):
         # Updating of the fact views should happen at the controller level,
@@ -1233,14 +1279,14 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
             criterion.card_type_deleted(card_type)
             self.update_criterion(criterion)
         del card_type
-        
+
     def has_card_type_with_id(self, id):
-        #if self.con.execute("select count() from card_types where id=?",
-        #    (id, )).fetchone()[0] != 0:
+        #if self.con.execute("select 1 from card_types where id=? limit 1",
+        #    (id, )).fetchone() is not None:
         #    return True
         #else: # It could be a built-in card type.
         return id in [card_type.id for card_type in self.card_types()]
-    
+
     #
     # Criteria.
     #
@@ -1300,11 +1346,11 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
     def criteria(self):
         return (self.criterion(cursor[0], is_id_internal=True) \
             for cursor in self.con.execute("select _id from criteria"))
-    
+
     def has_criterion_with_id(self, id):
-        return self.con.execute("select count() from criteria where id=?",
-            (id, )).fetchone()[0] != 0 
-    
+        return self.con.execute("select 1 from criteria where id=? limit 1",
+            (id, )).fetchone() is not None
+
 
     #
     # Queries.
@@ -1430,7 +1476,7 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
         card_type_2 = self.card_type_with_id("2")
         _fact_ids_dealt_with = []
         for key in set(_fact_id_for_front.keys()).\
-            intersection(_fact_id_for_back.keys()):
+            intersection(list(_fact_id_for_back.keys())):
             _fact_id_1 = _fact_id_for_front[key]
             _fact_id_2 = _fact_id_for_back[key]
             # Deal only once with a pair.
@@ -1507,9 +1553,10 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
     def cards_due_for_ret_rep(self, timestamp, sort_key="", limit=-1):
         sort_key = self._process_sort_key(sort_key)
         return ((cursor[0], cursor[1]) for cursor in self.con.execute("""
-            select _id, _fact_id from cards where
-            active=1 and grade>=2 and ?>=next_rep order by %s limit ?"""
-            % sort_key, (timestamp, limit)))
+            select _id, _fact_id from cards where active=1 and grade>=2
+            and ?>=next_rep and ret_reps_since_lapse<=? order by %s limit ?"""
+            % sort_key, (timestamp,
+            self.config()["max_ret_reps_since_lapse"], limit)))
 
     def cards_to_relearn(self, grade, sort_key="", limit=-1):
         sort_key = self._process_sort_key(sort_key)
@@ -1536,25 +1583,71 @@ _("Putting a database on a network drive is forbidden under Windows to avoid dat
         sort_key = self._process_sort_key(sort_key)
         return ((cursor[0], cursor[1]) for cursor in self.con.execute("""
             select _id, _fact_id from cards where
-            active=1 and grade>=2 and ?<next_rep order by %s limit ?"""
-            % sort_key, (timestamp, limit)))
+            active=1 and grade>=2 and ?<next_rep and ret_reps_since_lapse<=? 
+            order by %s limit ?"""
+            % sort_key, (timestamp,
+            self.config()["max_ret_reps_since_lapse"], limit)))
+
+    def recently_memorised_count(self, max_ret_reps):
+        return self.con.execute("""select count() from cards where active=1
+            and acq_reps>0 and ret_reps between 1 and ?""",
+            (max_ret_reps, )).fetchone()[0]
 
     #
-    # Extra commands for custom schedulers.
+    # Extra queries for custom schedulers.
     #
 
     def set_scheduler_data(self, scheduler_data):
         self.con.execute("update cards set scheduler_data=?",
             (scheduler_data, ))
 
-    def cards_with_scheduler_data(self, scheduler_data, sort_key="", limit=-1):
+    def cards_with_scheduler_data(self, scheduler_data, sort_key="",
+                                  limit=-1, max_ret_reps=-1):
         sort_key = self._process_sort_key(sort_key)
+        extra_cond = "" if max_ret_reps == -1 else str(
+            "and acq_reps>0 and ret_reps between 1 and " + str(max_ret_reps))
         return ((cursor[0], cursor[1]) for cursor in self.con.execute("""
-            select _id, _fact_id from cards where
-            active=1 and scheduler_data=? order by %s limit ?"""
-            % sort_key, (scheduler_data, limit)))
+            select _id, _fact_id from cards where active=1 and scheduler_data=?
+            %s order by %s limit ?"""
+            % (extra_cond, sort_key), (scheduler_data, limit)))
 
-    def scheduler_data_count(self, scheduler_data):
+    def scheduler_data_count(self, scheduler_data, max_ret_reps=-1):
+        extra_cond = "" if max_ret_reps == -1 else str(
+            "and acq_reps>0 and ret_reps between 1 and " + str(max_ret_reps))
         return self.con.execute("""select count() from cards
-            where active=1 and scheduler_data=?""",
+            where active=1 and scheduler_data=? %s """ % extra_cond,
             (scheduler_data, )).fetchone()[0]
+
+    def has_already_warned_today(self, start_of_day, end_of_day):
+        result = self.database().con.execute(
+            """select timestamp from log where ? <= log.timestamp and
+            log.timestamp <? and log.event_type=?""",
+            (start_of_day, end_of_day,
+             EventTypes.WARNED_TOO_MANY_CARDS)).fetchall()
+        return True if len(result) > 0 else False
+
+    #
+    # Extra queries for language analysis.
+    #
+
+    def _where_clause_known_recognition_questions(self, card_type_ids):
+        clause = "where grade>=2 and ( "
+        args = []
+        for card_type_id in card_type_ids:
+            clause += "(card_type_id=? and fact_view_id=?) or "
+            args += [card_type_id, card_type_id + ".1"]
+        clause = clause.rsplit("or ", 1)[0] + ")"
+        return clause, args
+
+    def known_recognition_questions_count_from_card_types_ids(\
+        self, card_type_ids):
+        clause, args = \
+            self._where_clause_known_recognition_questions(card_type_ids)
+        return self.con.execute(\
+            "select count() from cards " + clause, args).fetchone()[0]
+
+    def known_recognition_questions_from_card_types_ids(self, card_type_ids):
+        clause, args = \
+            self._where_clause_known_recognition_questions(card_type_ids)
+        return (cursor[0] for cursor in \
+                self.con.execute("select question from cards " + clause, args))
