@@ -2,7 +2,7 @@
 # scheduled_forgotten_new.py <Peter.Bienstman@UGent.be>
 #
 
-from mnemosyne.libmnemosyne.translator import _
+from mnemosyne.libmnemosyne.gui_translator import _
 from mnemosyne.libmnemosyne.study_mode import StudyMode
 from mnemosyne.libmnemosyne.schedulers.SM2_mnemosyne \
      import SM2Mnemosyne
