@@ -2,17 +2,16 @@
 # intro_wizard_dlg.py <Johannes.Baiter@gmail.com>
 #
 
-from PyQt4 import QtGui, QtCore
+from PyQt5 import QtGui, QtCore, QtWidgets
 
 from mnemosyne.pyqt_ui.ui_getting_started_dlg import Ui_GettingStartedDlg
 from mnemosyne.libmnemosyne.ui_components.dialogs import GettingStartedDialog
 
-class GettingStartedDlg(QtGui.QWizard, Ui_GettingStartedDlg,
-        GettingStartedDialog):
+class GettingStartedDlg(QtWidgets.QWizard, GettingStartedDialog, 
+                        Ui_GettingStartedDlg):
 
-    def __init__(self, component_manager):
-        GettingStartedDialog.__init__(self, component_manager)
-        QtGui.QWizard.__init__(self, self.main_widget())
+    def __init__(self, **kwds):
+        super().__init__(**kwds)
         self.setupUi(self)
         self.setWindowFlags(self.windowFlags() \
             | QtCore.Qt.WindowMinMaxButtonsHint)
@@ -22,13 +21,13 @@ class GettingStartedDlg(QtGui.QWizard, Ui_GettingStartedDlg,
         #watermark = QtGui.QPixmap(":/mnemosyne/pixmaps/mnemosyne.svg")\
         #    .scaledToHeight(200, QtCore.Qt.SmoothTransformation)
         watermark = QtGui.QPixmap(":/mnemosyne/pixmaps/mnemosyne.png")
-        self.setPixmap(QtGui.QWizard.WatermarkPixmap, watermark)
+        self.setPixmap(QtWidgets.QWizard.WatermarkPixmap, watermark)
 
     def activate(self):
         GettingStartedDialog.activate(self)
-        self.show()
+        self.exec_()
 
     def accept(self):
         self.config()["upload_science_logs"] = self.upload_box.isChecked()
-        QtGui.QWizard.accept(self)
+        QtWidgets.QWizard.accept(self)
 
